--[[ SPELL REDUX
#Author: Expop [US-Ravencrest]
#Date: Nov-22-2011
#Description: Addon to manage buffs/debuffs in a more visually
               stimulating way. Borrowed ideas from ShadowTimers
               addon. 
#License: GPLv3		
	Copyright (C) 2011 Dimitar Zlatkov 
	
	This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

]]--

-- ############### CONFIGURATION ###############
SpellRedux = {}
frameScale = 1
SPELL_SLOT_SIZE = 40;
framePositionsDefault = { SelfBuffFrame = {"CENTER", nil, "CENTER", 0, 0}; TargetDeBuffFrame = {"CENTER", nil, "CENTER", 0, SPELL_SLOT_SIZE} }
framePositions = framePositionsDefault;
MyAddon_UpdateInterval = 0.05;

targetDebuffSpells = {}
selfBuffSpells = {}

SlashCmdList["SPRED"] = SpellRedux_CommandLine;
SLASH_SPRED1 = "/SPRED";
	
hasFullyLoaded = false;
spells = {}
configMode = false;
classSpecs = {}
version = 0;
version_internal = 107
-- ############### CONFIGURATION ###############

-- Called when the addon is being loaded
function SpellRedux:OnLoad(frame)

	TimeSinceLastUpdate = 0	
	frame:RegisterEvent("ADDON_LOADED");
	frame:RegisterEvent("PLAYER_ALIVE");
	frame:RegisterEvent("ACTIVE_TALENT_GROUP_CHANGED");
	frame:RegisterEvent("PLAYER_REGEN_DISABLED");
	frame:RegisterEvent("PLAYER_REGEN_ENABLED");
	frame:RegisterEvent("PLAYER_TALENT_UPDATE");
	
	SelfBuffFrame:RegisterForDrag("LeftButton", "RightButton")
	SelfBuffFrame:EnableMouse(false);
	TargetDeBuffFrame:RegisterForDrag("LeftButton", "RightButton")
	TargetDeBuffFrame:EnableMouse(false);
	
	--SpellRedux_CommandLine("config")
end

-- Preloads spell defaults for this class/spec
function SpellRedux:preloadSpells()
	local lclass, class = UnitClass("player");
	classSpecs = table.copy(classes[class]);
	
	-- print("[SpellRedux]: Preloading defaults for",class);		
end

-- Loads the spells for the current class spec
function SpellRedux:loadClassSpec()
	local spec = GetSpecialization();
	
	if spec and #classSpecs ~= 0 then	
		-- print("[SpellRedux]: Loading class spec #",spec);
		targetDebuffSpells = classSpecs[spec].targetDebuffSpells;
		selfBuffSpells = classSpecs[spec].selfBuffSpells;		
	end	
	return spec
end

function SpellRedux:hasSpecChanged()
	local spec = GetSpecialization();
	return spec ~= nil and targetDebuffSpells ~= classSpecs[spec].targetDebuffSpells;
end

-- Called once per frame drawing
function SpellRedux:OnUpdate(elapsed)
		TimeSinceLastUpdate = TimeSinceLastUpdate + elapsed; 	
		while (TimeSinceLastUpdate > MyAddon_UpdateInterval) do
			SpellRedux:CheckCurrentTargetDeBuffs();
			SpellRedux:CheckPlayerBuffs();
			TimeSinceLastUpdate = TimeSinceLastUpdate - MyAddon_UpdateInterval;
		end
end

-- Configures a texture inside the given frame, with the set name, icon, dimensions and offsets
function SpellRedux:configTexture(frame, name, icon, w, h, ofsx, ofsy)
		local texture = SpellRedux:findObject(frame, "TOPLEFT", frame, "TOPLEFT", ofsx, ofsy, "Texture" );
		if texture == nil then
			texture = frame:CreateTexture(name .. "-tx");
			texture:SetWidth(w);
			texture:SetHeight(h);
			texture:SetDrawLayer("ARTWORK");
			texture:SetPoint("TOPLEFT", ofsx, ofsy);		
		end
		texture:SetTexture(icon);
		texture:SetAlpha(SPELL_OPACITY);
		texture:Show();
		return texture;
end

-- Finds an object if one can't be found with the given coordinates
function SpellRedux:findObject(frame, point, relativeTo, relativePoint, ofsx, ofsy, objtype )
	local regions =  {frame:GetRegions()};
	for i,t in ipairs(regions) do
		if t:GetObjectType() == objtype then
			local pt, rt, rp, xOfs, yOfs = t:GetPoint(1);
			if pt == point and rt == relativeTo and rp == relativePoint and areAlmostEqual(xOfs,ofsx) and areAlmostEqual(yOfs,ofsy) then
				return t;
			end
		end
	end
	return nil;
end

-- Hides all textures and fonts with location greater than x pixels 
function SpellRedux:hideObjects(frame, x )
	local regions =  {frame:GetRegions()};
	for i,t in ipairs(regions) do
		local tleft = t:GetLeft();
		if (areAlmostEqual(tleft,x) or tleft > x) and t:GetName() ~= nil then
			t:Hide();
		end
	end
	return nil;
end
--Accounts for jitter
function areAlmostEqual(x,y)
	local d = x - y 
	return d < 0.1 and d > -0.1
end

-- Creates the font in a given frame, and sets it with relation to the given UIObject
-- This is specifically meant to be used for the display of time
function SpellRedux:configFontTime(frame, name, relativeTo)
	local font = SpellRedux:findObject(frame, "CENTER", relativeTo, "CENTER", 0, 0, "FontString" );
	
	if font == nil then
		font = frame:CreateFontString(name .. "-ft", "OVERLAY", "GameFontNormalHuge");
		font:SetPoint("CENTER", relativeTo, "CENTER", 0, 0);
	end
	font:Show();
	return font;
end


-- Loads the textures for each spell in the spellList
-- for format of spellList, please see DisplaySpells.lua
function SpellRedux:loadTextrues(frame, spellList)
	local texture;
	local timefont;
	local multitex = {};
	local ofsx = 0;
	for i, spell_prop in ipairs(spellList) do	
		if spell_prop.prototype == nil then		
			local spell_name = spell_prop.spell_name;
			local name, rank, icon;
			if spell_prop.spell_id ~= nil then
				name, rank, icon = GetSpellInfo(spell_prop.spell_id);
			else
				name, rank, icon = GetSpellInfo(spell_name);
			end
			
			spell_prop.icon = icon;
			texture = SpellRedux:configTexture(frame,spell_name, icon, SPELL_SLOT_SIZE-1, SPELL_SLOT_SIZE-1, ofsx, -1);
			timefont = SpellRedux:configFontTime(frame, spell_name, texture);
			spells[spell_name] = Spell:new(spell_name,spell_prop.spell_id,spell_prop,texture,timefont)
			spells[spell_name]:Hide();
			ofsx = ofsx + SPELL_SLOT_SIZE;				
		end 
	end	
	SpellRedux:hideObjects(frame, frame:GetLeft() + ofsx);
	frame:SetWidth(ofsx);
end

function SpellRedux:linkSpells(spellList)
		for i, spell_prop in ipairs(spellList) do
			if spell_prop.prototype == PROTO_LINKABLE then
				local spell_name = spell_prop.spell_name;
				local spell_id = spell_prop.spell_id;				
				local name, rank, icon = GetSpellInfo(spell_name);
				spell_prop.icon = icon;
				for j, spell_s in ipairs(spellList) do
					if spell_s.spell_id == spell_prop.link_to then
						local linkedSpell = spells[spell_s.spell_name];
						local point, relativeTo, relativePoint, ofsx, ofsy = linkedSpell.texture:GetPoint(0);
						local frame = linkedSpell.texture:GetParent();
						texture = SpellRedux:configTexture(frame,spell_name, icon, SPELL_SLOT_SIZE-1, SPELL_SLOT_SIZE-1, ofsx, ofsy);
						timefont = SpellRedux:configFontTime(frame, spell_name, texture);
						spells[spell_name] = Spell:new(spell_name,spell_id,spell_prop,texture, timefont, linkedSpell)
						linkedSpell.linkedSpell = linkedSpell; --circular ref
						spells[spell_name]:Hide();
						break;
					end
				end				
			end
		end
end


-- Called when events are received by the addon
function SpellRedux:OnEvent(self,event, ...)
		local arg1 = ...		
		if (event == "ADDON_LOADED" and arg1 == "SpellRedux") then
 			if frameScale then
				SelfBuffFrame:SetScale(frameScale);
				TargetDeBuffFrame:SetScale(frameScale);
			end
			
			SpellRedux:CheckVersions();
			SpellRedux:UpdateSpellNames();
			
			-- FRAME POSITIONING			
			SelfBuffFrame:SetPoint(framePositions["SelfBuffFrame"][1], framePositions["SelfBuffFrame"][2], framePositions["SelfBuffFrame"][3], framePositions["SelfBuffFrame"][4], framePositions["SelfBuffFrame"][5]);
			TargetDeBuffFrame:SetPoint(framePositions["TargetDeBuffFrame"][1], framePositions["TargetDeBuffFrame"][2], framePositions["TargetDeBuffFrame"][3], framePositions["TargetDeBuffFrame"][4], framePositions["TargetDeBuffFrame"][5]);			
		elseif event == "ACTIVE_TALENT_GROUP_CHANGED" or event == "PLAYER_TALENT_UPDATE" then
			--ACTIVE_TALENT_GROUP_CHANGED used for when player switches talent specs
			--PLAYER_TALENT_UPDATE for when a user unlearns talents and picks new ones
			if not hasFullyLoaded or SpellRedux:hasSpecChanged() then
				hasFullyLoaded = true;
				SpellRedux:LoadBars(false);
			end
		elseif event == "PLAYER_REGEN_DISABLED" then
			SpellRedux:updateSpellActions();
			self:SetScript("OnUpdate", SpellRedux.OnUpdate);
			TargetDeBuffFrame:Show();
			SelfBuffFrame:Show();	
		elseif event == "PLAYER_REGEN_ENABLED" then
			self:SetScript("OnUpdate", nil);
			TargetDeBuffFrame:Hide();
			SelfBuffFrame:Hide();
			SpellRedux:turnShadingOff();
		end

end

function SpellRedux:turnShadingOff() 
	for _,spell in pairs(spells) do
		spell:removeShading();
	end
end

function SpellRedux:updateSpellActions()
	-- print('Loading spell actions');
	for _,spell in pairs(spells) do
		if spell.disable_glow == false then
			for i=1,120 do
				local t,id = GetActionInfo(i);
				if id ~= nil and t == 'spell' then
					local spell_name = GetSpellInfo(id);
					if id == spell.id or spell_name == spell.name then
						-- print('mapped ' .. spell_name .. ' => ' .. i);
						spell:setActionButtonId(i,id);
					end
					if id == spell.props.applied_by_id or spell_name == spell.props.applied_by then
						-- print('mapped ' .. spell_name .. ' => ' .. i);
						spell:setActionButtonId(i,id);
						break;
					end
				end
			end
		end
	end
end

-- combines multiple tables in this format [ [table1 elements], [table2 elements] ]
-- returns [table 1 elements, table 2 elements]
function combineTables( pt )
	local t = {};
	
	for i, v in ipairs(pt) do
		for j, e in ipairs(v) do
			table.insert(t, e)
		end
	end
	
	return t;
end

-- check if the addon has been upgraded 
function SpellRedux:CheckVersions()
	if version < version_internal then
		if version > 0 then
			--this is an upgrade
		else
			if #classSpecs == 0 then
				--this is a brand new installation
			else
				--pre-versioning installation 
				classSpecs = {}; --reset class specs
			end
		end
	end
	
	version = version_internal

end

-- update the spell names in the default list and the 
-- saved list too
function SpellRedux:UpdateSpellNames()
	for _, class in pairs(classes) do
		for i,spec in ipairs(class) do
			updateSpellNamesInList(spec.targetDebuffSpells);
			updateSpellNamesInList(spec.selfBuffSpells);
		end
	end
	
	for _, spec in ipairs(classSpecs) do
		updateSpellNamesInList(spec.targetDebuffSpells);
		updateSpellNamesInList(spec.selfBuffSpells);
	end
end
-- update the spell names in a spell list
function updateSpellNamesInList(spellList)
	local name = nil;
	for i,spell in ipairs(spellList) do
		if spell.spell_id then
			name = GetSpellInfo(spell.spell_id);
			if name == nil then
				print("Error loading spell " .. spell.spell_name);
				break;
			end
			spell.spell_name = name;		
		elseif spell.spell_name then 
			local spell_id = select(7,GetSpellInfo(spell.spell_name));
			if spell_id then
				spell.spell_id = spell_id
			end
		end
		if spell.applied_by_id then
			name = GetSpellInfo(spell.applied_by_id);
			spell.applied_by = name;
		elseif spell.applied_by then
			local spell_id = select(7,GetSpellInfo(spell.applied_by));
			if spell_id then 
				spell.applied_by_id = spell_id;
			end
		end
	end
end

-- Main loop which checks whether the given target has one of the spells from the 
-- spelllist, and if it does it updates the duration/stacks/shading and visibility 
-- of that spell. I've taken out the var declarations to make this perform with the 
-- amount of trash being generated. 
local displayList = {}
local spell_name;
local bn,brank,bicon,bcount,bType,bduration,bexpirationTime,bisMine;
function SpellRedux:checkSpells(target, filter, spelllist)
	for i, spell in ipairs(spelllist) do
		spell_name = spell.spell_name;
		bn,brank,bicon,bcount,bType,bduration,bexpirationTime,bisMine =  UnitAura(target, spell_name, nil, filter )		
		if bn ~= nil then			
			if spells[spell_name].texture:GetTexture() == nil then
				spells[spell_name].texture:SetTexture( bicon );
				spells[spell_name].props.icon = bicon;
			end		
			spells[spell_name]:refresh(bduration, (bexpirationTime-GetTime()), bcount);
			displayList[spell_name] = true;
		end
	end
	
	
	for i, spell in ipairs(spelllist) do
		spell_name = spell.spell_name;			
		if displayList[spell_name] then			
			if not spells[spell_name]:IsShown() then
				spells[spell_name]:Show();
			end
			displayList[spell_name] = false;
		elseif not displayList[spell_name] then			
			if spells[spell_name]:IsShown() then spells[spell_name]:Hide(); end
			if spells[spell_name].disable_glow == false then spells[spell_name].shader:glow(true); end
		end
	end
	
end

function SpellRedux:CheckCurrentTargetDeBuffs()
	SpellRedux:checkSpells("target", "PLAYER|HARMFUL", targetDebuffSpells);
end

function SpellRedux:CheckPlayerBuffs()
	SpellRedux:checkSpells("player", "PLAYER|HELPFUL", selfBuffSpells);
end

-- for moving the bars
function SpellRedux:StartMoveFrame(self, button)		
	if configMode then
		if button == "LeftButton" and not self.isLocked then
		self:StartMoving();
		self.isMoving = true;
		end
	end
end

-- for moving the bars
function SpellRedux:StopMoveFrame(self, button)	
	if button == "LeftButton" and self.isMoving then
		self:StopMovingOrSizing();
		self.isMoving = false;
		point, relativeTo, relativePoint, xOffset, yOffset = self:GetPoint(1);
		framePositions[self:GetName()][1] = point;
		framePositions[self:GetName()][2] = relativeTo;
		framePositions[self:GetName()][3] = relativePoint;
		framePositions[self:GetName()][4] = xOffset;
		framePositions[self:GetName()][5] = yOffset;
		
	end
end

function SpellRedux_CommandLine(msg)	
	if InCombatLockdown() == 1 then
		print("Can't modify addons while in combat.");
		return;
	end 
	if  msg == 'lock' then	
		if SelfBuffFrame:GetWidth() > 0 then
			SelfBuffFrame:Hide();
			SelfBuffFrame:EnableMouse(false);
			SelfBuffFrame:SetBackdrop(nil);
		end
		if TargetDeBuffFrame:GetWidth() > 0 then
			TargetDeBuffFrame:Hide();
			TargetDeBuffFrame:EnableMouse(false);
			TargetDeBuffFrame:SetBackdrop(nil);
		end
		configMode = false;
	elseif  msg == 'unlock' then
		if SelfBuffFrame:GetWidth() > 0 then
			SelfBuffFrame:Show();
			SelfBuffFrame:EnableMouse(true);
			SelfBuffFrame:SetBackdrop({bgFile = "Interface/Tooltips/UI-Tooltip-Background", edgeFile= "Interface/Tooltips/UI-Tooltip-Border", edgeSize = 4, tile = false, tileSize =16, insets = { left = 0, right = 0, top = 0, bottom = 0 }});
		end
		if TargetDeBuffFrame:GetWidth() > 0 then
			TargetDeBuffFrame:Show();
			TargetDeBuffFrame:EnableMouse(true);
			TargetDeBuffFrame:SetBackdrop({bgFile = "Interface/Tooltips/UI-Tooltip-Background", edgeFile= "Interface/Tooltips/UI-Tooltip-Border", edgeSize = 4, tile = false, tileSize =16, insets = { left = 0, right = 0, top = 0, bottom = 0 }});
		end	
		configMode = true;
	elseif strfind(msg, 'scale') ~= nil then
		local scale = tonumber(strsub(msg,strfind(msg, 'scale') +5));
		if scale ~= nil and scale > 0 and scale <= 10 then
			if SelfBuffFrame:GetWidth() > 0 then
				SelfBuffFrame:SetScale(scale/10);
			end
			if TargetDeBuffFrame:GetWidth() > 0 then
				TargetDeBuffFrame:SetScale(scale/10);
			end
		else
			print ("Bad scale range. Specify number from 1 to 10 inclusive, 1 being tiny and 10 being full-sized");
		end
	elseif msg == 'reset' then
		SelfBuffFrame:SetScale(1.0);
		TargetDeBuffFrame:SetScale(1.0);
		SelfBuffFrame:SetPoint("CENTER", nil, "CENTER", 0, 0);
		TargetDeBuffFrame:SetPoint("CENTER", nil, "CENTER", 0, 0);
		framePositions = framePositionsDefault;
	elseif msg == 'defaults' then	
		for k,spell in pairs (spells) do
			spell:Hide();
		end		
		SpellRedux:LoadBars(true);
		for k,spell in ipairs (selfBuffSpells) do
		 print("Loaded:", spell.spell_name);
		end
		for k,spell in ipairs (targetDebuffSpells) do
		 print("Loaded:", spell.spell_name);
		end
	elseif msg == 'config' then
		InterfaceOptionsFrame_OpenToCategory("SpellRedux");
		InterfaceOptionsFrame_OpenToCategory("Target Debuffs"); -- blizz fix
	else
	   print("Syntax: /spred \nlock - locks the bars\nunlock - unlock so bars can be moved\nreset - reset bar positions",
			 "\ndefaults - get the default spells for your spec",
			 "\nconfig - opens the configuration GUI",
	         "\nscale <1-10> - scale the bars (1 smallest, 10 biggest). Normal size is 10. ");

	 end
end

function SpellRedux:LoadBars(forceDefaults)
	if #classSpecs == 0 or forceDefaults then
		spells = {}
		SpellRedux:preloadSpells();
	end
	if SpellRedux:loadClassSpec() then
		SpellRedux:loadTextrues(SelfBuffFrame,selfBuffSpells);
		SpellRedux:loadTextrues(TargetDeBuffFrame,targetDebuffSpells);				
		SpellRedux:linkSpells( combineTables( {selfBuffSpells, targetDebuffSpells} ));
	end
	SpellRedux:updateSpellActions();
end

-- where spell is an array. see DisplaySpells for example.
function SpellRedux:NewSpell(newSpell, spellList)
	if InCombatLockdown() == 1 then
		print("Can't modify addons while in combat.");
		return;
	end 
	new = true;
	for i,spell in ipairs(spellList) do
		if spell.spell_name == newSpell.spell_name then
			spellList[i] = newSpell;
			new = false;
			break;
		end
	end
	if new then 
		table.insert(spellList, newSpell);
	end
	SpellRedux:LoadBars(false);
end

-- where spellName is the name of the spell (e.g. Immolate)
function SpellRedux:DeleteSpell(spellName, spellList)
	if InCombatLockdown() == 1 then
		print("Can't modify addons while in combat.");
		return;
	end 
	for k,spell in pairs (spells) do
		spell:Hide();
	end
	spells = {};
	local deletedSpell = nil;
	
	for i,spell in ipairs(spellList) do
		if spell.spell_name == spellName then
			deletedSpell = spell;
			table.remove(spellList, i);
			break;
		end
	end
	-- update linked spells
	if deletedSpell ~= nil then
		for i,spell in ipairs(spellList) do
			if spell.link_to == deletedSpell.spell_id then
				spell.prototype = nil;
				spell.link_to = nil;
			end
		end
	end
	SpellRedux:LoadBars(false);
end

-- spell - id or name
function GetSpell(spellList, spell)
	for _,sp in pairs(spellList) do
		if sp.spell_name == spell or sp.spell_id == spell then
			return sp;
		end
	end
	return nil;
end

-- deep-copies a table
-- does not copy userdata and thread
function table.copy(base)
	local new = {};
	for k,v in pairs(base) do
		local t = type(v);
		
		if t == nil or t == "boolean" or t == "number" or t == "string" or t == "function" then
			new[k] = v;
		elseif t == "table"  then
			new[k] = table.copy(v);
		end
	end
	return new;
end
